/*
 * Project: VizKit
 * Version: 3.3.6
 
 * Date: 20150404
 * File: TemplateActor.cpp
 *
 */

/***************************************************************************

Copyright (c) 2004-2015 Heiko Wichmann (http://www.imagomat.de/vizkit)


This software is provided 'as-is', without any expressed or implied warranty. 
In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; 
   you must not claim that you wrote the original software. 
   If you use this software in a product, an acknowledgment 
   in the product documentation would be appreciated 
   but is not required.

2. Altered source versions must be plainly marked as such, 
   and must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source distribution.

 ***************************************************************************/

#include "TemplateActor.h"
#include "TemplateAction.h"
#include "VisualErrorHandling.h"
#include "VisualString.h"
#include "VisualStyledString.h"
#include "VisualNetwork.h"
#include "VisualDataStore.h"
#include "VisualFile.h"
#include "VisualNotification.h"


#if TARGET_OS_MAC
#include <Carbon/Carbon.h> // CFReadStreamRef, CFURLRef
#endif

#include <string>



using namespace VizKit;


TemplateActor::TemplateActor() {
	strcpy(actorName, "TEMPLATE");
	state = kVisActOn; // state must be kVisActOn to show the action of the actor as part of the VisualEnsemble (kVisActNoShow means not being called at all)
	theTemplateAction = new TemplateAction;
	stringOfSimilarArtists = NULL;
	similarArtistsStringCreationInProgress = false;
	this->createStringOfSimilarArtists();
}


TemplateActor::~TemplateActor() {
    delete theTemplateAction;
	similarArtistsStringVector.clear();
	if (stringOfSimilarArtists != NULL) {
		delete stringOfSimilarArtists;
	}
}


void TemplateActor::init() {
	VisualNotification::registerNotification(this, kStringWithIdentifierLoadedAndCreatedMsg);
	VisualNotification::registerNotification(this, kAudioPlayTrackChangedEvt);
    VisualNotification::registerNotification(this, kCanvasReshapeEvt);
}


void TemplateActor::show(const VisualPlayerState& visualPlayerState) {
    theTemplateAction->show();
}


char * replace(
               char const * const original,
               char const * const pattern,
               char const * const replacement
               ) {
    
    if ((original != NULL) && ((pattern == NULL) || (replacement == NULL)))
    {
        char * const returnedOriginalCopy = (char *) malloc( sizeof(char) * (strlen(original) + 1) );
        strncpy(returnedOriginalCopy, original, strlen(original));
        return returnedOriginalCopy;
    }
    else if ((original == NULL) || (pattern == NULL) || (replacement == NULL))
    {
        return NULL;
    }
    
    size_t const replen = strlen(replacement);
    size_t const patlen = strlen(pattern);
    size_t const orilen = strlen(original);
    
    size_t patcnt = 0;
    const char * oriptr;
    const char * patloc;
    
    // find how many times the pattern occurs in the original string
    for (oriptr = original; (patloc = strstr(oriptr, pattern)); oriptr = patloc + patlen)
    {
        patcnt++;
    }
    
    {
        // allocate memory for the new string
        size_t const retlen = orilen + patcnt * (replen - patlen);
        char * const returned = (char *) malloc( sizeof(char) * (retlen + 1) );
        
        if (returned != NULL)
        {
            // copy the original string,
            // replacing all the instances of the pattern
            char * retptr = returned;
            for (oriptr = original; (patloc = strstr(oriptr, pattern)); oriptr = patloc + patlen)
            {
                size_t const skplen = patloc - oriptr;
                // copy the section until the occurence of the pattern
                strncpy(retptr, oriptr, skplen);
                retptr += skplen;
                // copy the replacement
                strncpy(retptr, replacement, replen);
                retptr += replen;
            }
            // copy the rest of the string.
            strcpy(retptr, oriptr);
        }
        return returned;
    }
}

void TemplateActor::handleSimilarArtistsStringCompleted() {

	if (this->stringOfSimilarArtists == NULL) return;
	
	VisualItemIdentifier currTrackIdentifier = VisualDataStore::getIdentifierOfCurrentTrack();
	if (currTrackIdentifier != this->similarArtistsCurrentTrackIdentifier) {
		return;
	}
	
	uint32 numberOfLines = this->stringOfSimilarArtists->getNumberOfLines();

	for (uint32 i = 0; i < numberOfLines; i++) {
		VisualString* visualString = this->stringOfSimilarArtists->createStringWithLine(i);
		if (visualString) {
			std::vector<VisualString> stringVector = visualString->splitWithDelimiter(",");
			if (stringVector.size() > 2) {
				this->similarArtistsStringVector.push_back(stringVector[2]);
			}
			delete visualString;
		}
	}
    
    if (similarArtistsStringVector.empty())
    {
        return;
    }
	
	VisualString* joinedString = VisualString::createJoinedString(this->similarArtistsStringVector);
    const char* const joinedStringUtf8 = joinedString->getUtf8Representation();
    char* joinedStringUtf8Mutable = (char*)malloc(strlen(joinedStringUtf8) * sizeof(char) + 1);
    strncpy(joinedStringUtf8Mutable, joinedStringUtf8, strlen(joinedStringUtf8));
	joinedStringUtf8Mutable[strlen(joinedStringUtf8)] = '\0';
    char *joinedStringUtf8WithHtmlSpecialCharReplacedOne = replace(joinedStringUtf8Mutable, "&amp;", "&");
    char *joinedStringUtf8WithHtmlSpecialCharReplacedTwo = replace(joinedStringUtf8WithHtmlSpecialCharReplacedOne, "&quot;", "\"");
    VisualString *joinedStringUtf8WithHtmlSpecialCharReplacedFinal = VisualString::createWithUTF8String(joinedStringUtf8WithHtmlSpecialCharReplacedTwo);
	if (joinedStringUtf8WithHtmlSpecialCharReplacedFinal != NULL) {
		this->setStringOfSimilarArtists(*joinedStringUtf8WithHtmlSpecialCharReplacedFinal);
		delete joinedStringUtf8WithHtmlSpecialCharReplacedFinal;
	}
    if (joinedString)
    {
        delete joinedString;
    }
    if (joinedStringUtf8Mutable)
    {
        free(joinedStringUtf8Mutable);
    }
    if (joinedStringUtf8WithHtmlSpecialCharReplacedOne)
    {
        free(joinedStringUtf8WithHtmlSpecialCharReplacedOne);
    }
    if (joinedStringUtf8WithHtmlSpecialCharReplacedTwo)
    {
        free(joinedStringUtf8WithHtmlSpecialCharReplacedTwo);
    }
}


void TemplateActor::setStringOfSimilarArtists(const VisualString& similarArtists) {
	VisualStyledString styledString;
	styledString.initWithVisualString(similarArtists);
	VisualStringStyle stringStyle;
	stringStyle.horizontalAlignment = kRightAligned;
	styledString.setStyle(stringStyle);

	bool success = this->similarArtistsStringImage.initWithStyledString(styledString);
	if (success) {
		this->theTemplateAction->setImage(this->similarArtistsStringImage);
		/*
		VisualFile aFile;
		aFile.initWithUserDesktopDirectory();
		VisualString fileName("testSimilarArtists.png");
		success = aFile.appendFileName(fileName);
		this->similarArtistsStringImage.writeToPNGFile(aFile);
		*/
	}
}


void TemplateActor::handleNotification(VisualNotification& aNotification) {

	//VisualActor::handleNotification(aNotification); // debug

	VisualNotificationKey notificationKey = aNotification.getKey();
	
	switch (notificationKey) {
			
		case kStringWithIdentifierLoadedAndCreatedMsg:
			if (aNotification.getIdentifier() == this->similarArtistsStringRequestId) {
				this->handleSimilarArtistsStringCompleted();
			}
			break;
			
		case kAudioPlayTrackChangedEvt:
			this->createStringOfSimilarArtists();
			break;
            
        case kCanvasReshapeEvt:
			this->theTemplateAction->reshape();
			break;
			
		default:
			char notificationString[64];
			VisualNotification::convertNotificationKeyToString(notificationKey, notificationString);
			char errLog[512];
			sprintf(errLog, "Unhandled VisualNotificationKey %s in file: %s (line: %d) [%s])", notificationString, __FILE__, __LINE__, __FUNCTION__);
			writeLog(errLog);
			break;
	}

}


void TemplateActor::createStringOfSimilarArtists() {

	if (this->similarArtistsStringCreationInProgress == true) {
		return;
	}
	
	this->similarArtistsStringCreationInProgress = true;

	this->similarArtistsStringVector.clear();
	this->theTemplateAction->removeImage();
	
	this->similarArtistsCurrentTrackIdentifier = VisualDataStore::getIdentifierOfCurrentTrack();
	
	VisualString* normalizedArtistNameForURLConstruction = VisualNetwork::createNormalizedURLString(VisualDataStore::getArtistOfCurrentTrack());
	if (!normalizedArtistNameForURLConstruction) {
		return;
	}
		
	VisualString similarArtistsURL;
	similarArtistsURL.initWithUTF8String("http://ws.audioscrobbler.com/1.0/artist/");
	similarArtistsURL = (similarArtistsURL + *normalizedArtistNameForURLConstruction);
	delete normalizedArtistNameForURLConstruction;
	similarArtistsURL = (similarArtistsURL + "/similar.txt");
	if (this->stringOfSimilarArtists != NULL) {
        VisualNetwork::removeDownloadForObject(this->stringOfSimilarArtists->getIdentifier());
		delete this->stringOfSimilarArtists;
	}

	this->stringOfSimilarArtists = VisualString::createWithContentsOfURL(similarArtistsURL.getUtf8Representation(), this->similarArtistsStringRequestId);

	this->similarArtistsStringCreationInProgress = false;
}
