/*
 * Project: VizKit
 * Version: 3.3.6
 
 * Date: 20150404
 * File: VisualAudioData.cpp
 *
 */

/***************************************************************************

Copyright (c) 2004-2015 Heiko Wichmann (http://www.imagomat.de/vizkit)


This software is provided 'as-is', without any expressed or implied warranty. 
In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; 
   you must not claim that you wrote the original software. 
   If you use this software in a product, an acknowledgment 
   in the product documentation would be appreciated 
   but is not required.

2. Altered source versions must be plainly marked as such, 
   and must not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source distribution.

 ***************************************************************************/

#include "VisualAudioData.h"

#include <cstdlib>


using namespace VizKit;


VisualAudioData::VisualAudioData() {
    waveformChannelCount = 0;
    waveformEntryCount = 0;
    waveformData = NULL;
    spectrumChannelCount = 0;
    spectrumEntryCount = 0;
    spectrumData = NULL;
}


VisualAudioData::~VisualAudioData() {
	if (waveformData)
    {
        free(waveformData);
    }
    if (spectrumData)
    {
        free(spectrumData);
    }
}


VisualAudioData::VisualAudioData(const VisualAudioData& other) {
	this->copy(other);
}


VisualAudioData& VisualAudioData::operator=(const VisualAudioData& other) {
	
	if (this == &other) return *this;

	this->copy(other);

	return *this;
}


bool VisualAudioData::operator==(const VisualAudioData& other) {
	bool compareResult = true;
	if ((this->waveformChannelCount == other.waveformChannelCount) &&
		(this->waveformEntryCount == other.waveformEntryCount) &&
		(this->spectrumChannelCount == other.spectrumChannelCount) &&
		(this->spectrumEntryCount == other.spectrumEntryCount))
    {
        for (uint8 aWaveformChannel = 0; aWaveformChannel < this->waveformChannelCount; aWaveformChannel++) 
        {
            for (uint16 aWaveformEntryIndex = 0; aWaveformEntryIndex < this->waveformEntryCount; aWaveformEntryIndex++) 
            {
                if ((this->waveformData[aWaveformChannel * this->waveformEntryCount + aWaveformEntryIndex]) != (other.waveformData[aWaveformChannel * this->waveformEntryCount + aWaveformEntryIndex]))
                {
                    compareResult = false;
                    break;
                }
            }
            if (compareResult == false)
            {
                break;
            }
        }
        if (compareResult == true)
        {
            for (uint8 aSpectrumChannel = 0; aSpectrumChannel < this->spectrumChannelCount; aSpectrumChannel++) 
            {
                for (uint16 aSpectrumEntryIndex = 0; aSpectrumEntryIndex < this->spectrumEntryCount; aSpectrumEntryIndex++) 
                {
                    if ((this->spectrumData[aSpectrumChannel * this->spectrumEntryCount + aSpectrumEntryIndex]) != (other.spectrumData[aSpectrumChannel * this->spectrumEntryCount + aSpectrumEntryIndex]))
                    {
                        compareResult = false;
                        break;
                    }
                }
                if (compareResult == false)
                {
                    break;
                }
            }
        }
	}
	return compareResult;
}


bool VisualAudioData::operator!=(const VisualAudioData& other) {
	return !(*this == other);
}


void VisualAudioData::copy(const VisualAudioData& other) {
	this->waveformChannelCount = other.waveformChannelCount;
	this->waveformEntryCount = other.waveformEntryCount;
	this->spectrumChannelCount = other.spectrumChannelCount;
	this->spectrumEntryCount = other.spectrumEntryCount;
    
    if (this->waveformData)
    {
        free(this->waveformData);
    }
    this->waveformData = (uint8*)malloc(this->waveformChannelCount * this->waveformEntryCount * sizeof(uint8));
    if (this->spectrumData)
    {
        free(this->spectrumData);
    }
    this->spectrumData = (uint8*)malloc(this->spectrumChannelCount * this->spectrumEntryCount * sizeof(uint8));
    
    for (uint8 aWaveformChannel = 0; aWaveformChannel < this->waveformChannelCount; aWaveformChannel++) 
    {
        for (uint16 aWaveformEntryIndex = 0; aWaveformEntryIndex < this->waveformEntryCount; aWaveformEntryIndex++) 
        {
            this->waveformData[aWaveformChannel * this->waveformEntryCount + aWaveformEntryIndex] = other.waveformData[aWaveformChannel * this->waveformEntryCount + aWaveformEntryIndex];
        }
    }
    for (uint8 aSpectrumChannel = 0; aSpectrumChannel < this->spectrumChannelCount; aSpectrumChannel++) 
    {
        for (uint16 aSpectrumEntryIndex = 0; aSpectrumEntryIndex < this->spectrumEntryCount; aSpectrumEntryIndex++) 
        {
            this->spectrumData[aSpectrumChannel * this->spectrumEntryCount + aSpectrumEntryIndex] = other.spectrumData[aSpectrumChannel * this->spectrumEntryCount + aSpectrumEntryIndex];
        }
    }
}


void VisualAudioData::setNumberOfWaveformChannels(uint8 aWaveformChannelCount)
{
    this->waveformChannelCount = aWaveformChannelCount;
}


void VisualAudioData::setNumberOfWaveformEntries(uint16 aWaveformEntryCount)
{
    this->waveformEntryCount = aWaveformEntryCount;
}

	
void VisualAudioData::setWaveformData(const uint8* const someWaveformData)
{
    if (this->waveformData != NULL)
    {
        free(this->waveformData);
    }
    this->waveformData = (uint8*)malloc(this->waveformChannelCount * this->waveformEntryCount * sizeof(uint8));
    for (uint8 aWaveformChannel = 0; aWaveformChannel < this->waveformChannelCount; aWaveformChannel++) 
    {
        for (uint16 aWaveformEntryIndex = 0; aWaveformEntryIndex < this->waveformEntryCount; aWaveformEntryIndex++) 
        {
            this->waveformData[aWaveformChannel * this->waveformEntryCount + aWaveformEntryIndex] = someWaveformData[aWaveformChannel * this->waveformEntryCount + aWaveformEntryIndex];
        }
    }
}


void VisualAudioData::setNumberOfSpectrumChannels(uint8 aSpectrumChannelCount)
{
    this->spectrumChannelCount = aSpectrumChannelCount;
}


void VisualAudioData::setNumberOfSpectrumEntries(uint16 aSpectrumEntryCount)
{
    this->spectrumEntryCount = aSpectrumEntryCount;
}

	
void VisualAudioData::setSpectrumData(const uint8* const someSpectrumData)
{
    if (this->spectrumData != NULL)
    {
        free(this->spectrumData);
    }
    this->spectrumData = (uint8*)malloc(this->spectrumChannelCount * this->spectrumEntryCount * sizeof(uint8));
    for (uint8 aSpectrumChannel = 0; aSpectrumChannel < this->spectrumChannelCount; aSpectrumChannel++) 
    {
        for (uint16 aSpectrumEntryIndex = 0; aSpectrumEntryIndex < this->spectrumEntryCount; aSpectrumEntryIndex++) 
        {
            this->spectrumData[aSpectrumChannel * this->spectrumEntryCount + aSpectrumEntryIndex] = someSpectrumData[aSpectrumChannel * this->spectrumEntryCount + aSpectrumEntryIndex];
        }
    }
}


uint8 VisualAudioData::getNumberOfWaveformChannels() const
{
    return this->waveformChannelCount;
}


uint16 VisualAudioData::getNumberOfWaveformEntries() const
{
    return this->waveformEntryCount;
}


const uint8* const VisualAudioData::getWaveformData() const
{
    return this->waveformData;
}


uint8 VisualAudioData::getNumberOfSpectrumChannels() const
{
    return this->spectrumChannelCount;
}


uint16 VisualAudioData::getNumberOfSpectrumEntries() const
{
    return this->spectrumEntryCount;
}


const uint8* const VisualAudioData::getSpectrumData() const
{
    return this->spectrumData;
}
