/*
 * Project: VizKit
 * Version: 3.3.6
 
 * Date: 20150404
 * File: VisualView.m
 *
 */

/***************************************************************************
 
 Copyright (c) 2004-2015 Heiko Wichmann (http://www.imagomat.de/vizkit)
 
 
 This software is provided 'as-is', without any expressed or implied warranty. 
 In no event will the authors be held liable for any damages
 arising from the use of this software.
 
 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:
 
 1. The origin of this software must not be misrepresented; 
 you must not claim that you wrote the original software. 
 If you use this software in a product, an acknowledgment 
 in the product documentation would be appreciated 
 but is not required.
 
 2. Altered source versions must be plainly marked as such, 
 and must not be misrepresented as being the original software.
 
 3. This notice may not be removed or altered from any source distribution.
 
 ***************************************************************************/


#import "VisualView.h"
#include <OpenGL/OpenGL.h>
#include "VisualCustomization.h"
#include "VisualMainAction.h"


// file type of VisualView.m is set to "Objective-C++ source"


@interface VisualView()
- (void)drawFrame:(NSRect)aFrame;
@end



@implementation VisualView




- (id) initWithFrame:(NSRect)aFrame
{
    didHideCursor = NO;
    
    NSOpenGLPixelFormatAttribute pixelFormatAttributes[] =
	{
        NSOpenGLPFAWindow,
        NSOpenGLPFADoubleBuffer,
        NSOpenGLPFASingleRenderer,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFAAccelerated,
        0
	};
    
    NSOpenGLPixelFormat *aPixelFormat = [[[NSOpenGLPixelFormat alloc] initWithAttributes:pixelFormatAttributes] autorelease];
    
    if (self = [super initWithFrame:aFrame pixelFormat:aPixelFormat])
    {
        if ([self respondsToSelector:@selector(setWantsBestResolutionOpenGLSurface:)])
        {
            [self setWantsBestResolutionOpenGLSurface:YES];
        }
    }
    
    return self;
}


- (void)dealloc
{
    [self cleanup];
    
    [super dealloc];
}


- (BOOL) isOpaque
{
    return YES;
}


CVReturn MyDisplayLinkCallback(CVDisplayLinkRef displayLink, const CVTimeStamp* now, const CVTimeStamp* outputTime, CVOptionFlags flagsIn, CVOptionFlags* flagsOut, void* displayLinkContext)
{
    CVReturn result = [(VisualView*)displayLinkContext getFrameForTime:outputTime];
    return result;
}


- (CVReturn)getFrameForTime:(const CVTimeStamp*)outputTime
{
    @autoreleasepool
    {
        if ([self lockFocusIfCanDraw])
        {
            [self drawFrame:self.bounds];
            [self unlockFocus];
        }
    }
    return kCVReturnSuccess;
}


- (void)drawFrame:(NSRect)aFrame
{
    // this is called in worker thread
    NSAutoreleasePool *aPool = [[NSAutoreleasePool alloc] init];
    
    NSOpenGLContext *aCurrentContext = [self openGLContext];
    if (aCurrentContext)
    {
        [aCurrentContext makeCurrentContext];
        
        // This method will be called on both the main thread (through -drawRect:) and a secondary thread (through the display link rendering loop)
        // Also, when resizing the view, -reshape is called on the main thread, but we may be drawing on a secondary thread
        // Add a mutex around to avoid the threads accessing the context simultaneously
        CGLLockContext((CGLContextObj)[aCurrentContext CGLContextObj]);
        
        VizKit::VisualMainAction::renderAction();
        
        [aCurrentContext flushBuffer]; // calls through to CGLFlushDrawable()
        
        CGLUnlockContext((CGLContextObj)[aCurrentContext CGLContextObj]);
    }
    
    [aPool drain];
    
    return kCVReturnSuccess;
}


- (void)setup
{
    NSOpenGLContext *aCurrentContext = [self openGLContext];
    if (aCurrentContext)
    {
        [aCurrentContext makeCurrentContext];
        
        GLint aSwapInterval = 1;
        [aCurrentContext setValues:&aSwapInterval forParameter:NSOpenGLCPSwapInterval];
        
        // create a display link capable of being used with all active displays
        CVDisplayLinkCreateWithActiveCGDisplays(&displayLink);
        
        // set the renderer output callback function
        CVDisplayLinkSetOutputCallback(displayLink, &MyDisplayLinkCallback, self);
        
        // set the display link for the current renderer
        CGLContextObj cglContext = (CGLContextObj)[aCurrentContext CGLContextObj];
        CGLPixelFormatObj cglPixelFormat = (CGLPixelFormatObj)[[self pixelFormat] CGLPixelFormatObj];
        CVDisplayLinkSetCurrentCGDisplayFromOpenGLContext(displayLink, cglContext, cglPixelFormat);
    }
}


- (void)start
{
    CVDisplayLinkStart(displayLink);
}


- (void)cleanup
{
    NSOpenGLContext *aCurrentContext = [self openGLContext];
    if (aCurrentContext)
    {
        CGLLockContext((CGLContextObj)[aCurrentContext CGLContextObj]);
        
        if (displayLink != NULL)
        {
            if (didHideCursor)
            {
                CGDisplayShowCursor(CVDisplayLinkGetCurrentCGDisplay(displayLink));
            }
            if (CVDisplayLinkIsRunning(displayLink))
            {
                CVDisplayLinkStop(displayLink);
            }
            CVDisplayLinkRelease(displayLink);
            displayLink = NULL;
        }
        
        CGLUnlockContext((CGLContextObj)[aCurrentContext CGLContextObj]);
    }
}


- (void)update
{
    NSOpenGLContext *aCurrentContext = [self openGLContext];
    if (aCurrentContext)
    {
        CGLLockContext((CGLContextObj)[aCurrentContext CGLContextObj]);
        [aCurrentContext makeCurrentContext];
        [aCurrentContext update];
        CGLUnlockContext((CGLContextObj)[aCurrentContext CGLContextObj]);
    }
}


- (void)reshape
{
    
    if (didHideCursor == NO)
    {
        if ([NSScreen screens].count == 1)
        {
            NSRect screenFrame = [[NSScreen mainScreen] frame];
            if ((screenFrame.size.width == self.frame.size.width) && (screenFrame.size.height == self.frame.size.height))
            {
                CGDisplayHideCursor(CVDisplayLinkGetCurrentCGDisplay(displayLink));
                didHideCursor = YES;
            }
        }
    }
    
    if (displayLink != NULL)
    {
        NSOpenGLContext *aCurrentContext = [self openGLContext];
        if (aCurrentContext)
        {
            CGLLockContext((CGLContextObj)[aCurrentContext CGLContextObj]);
            [aCurrentContext update];
            CGLUnlockContext((CGLContextObj)[aCurrentContext CGLContextObj]);
            VizKit::VisualMainAction::handleCanvasReshape();
            if ([self lockFocusIfCanDraw])
            {
                [self drawFrame:self.bounds];
                [self unlockFocus];
            }            
        }
    }
}


- (void)drawRect:(NSRect)dirtyRect
{
    if (!CVDisplayLinkIsRunning(displayLink))
    {
        [self drawFrame:self.bounds];
    }
}


#pragma mark - Event Handling


- (BOOL)acceptsFirstResponder
{
    return YES;
}


- (void)keyDown:(NSEvent *)anEvent
{
    bool handled = false;
    
    NSString *characters = [anEvent charactersIgnoringModifiers];
    NSString *aCharString = [characters substringToIndex:1];
    const char *aChar = [aCharString UTF8String];
    if (strlen(aChar) == 1)
    {
        handled = VizKit::VisualMainAction::handleKeyPressEvent(*aChar);
    }
    
    if (handled == false)
    {
        [super keyDown:anEvent];
    }
}


@end
